# standard library
import shutil
import tomllib

from glob import glob
from pathlib import Path

# typing
from typing import Callable

# third parties
from pydantic import BaseModel

# Youwol
import youwol

# Youwol application
from youwol.app.environment import YouwolEnvironment
from youwol.app.routers.projects import (
    Artifact,
    ExplicitNone,
    Family,
    FileListing,
    Flow,
    Pipeline,
    PipelineStep,
    Project,
    RunImplicit,
    Target,
)

# Youwol utilities
from youwol.utils import AnyDict, Context, execute_shell_cmd, write_json

# Youwol pipelines
from youwol.pipelines import (
    Environment,
    PublishCdnLocalStep,
    create_sub_pipelines_publish_cdn,
)


class Dependencies:
    get_environment: Callable[[], Environment] = Environment


def set_environment(environment: Environment = Environment()):
    """
    The environment for this pipeline is only defining the remote ecosystem(s).

    Parameters:
        environment: The environment
    """
    Dependencies.get_environment = lambda: environment


def get_environment() -> Environment:
    return Dependencies.get_environment()


pyproject_file = "pyproject.toml"


class SetupStep(PipelineStep):
    """
    Initializes auto-generated files from `pyproject.toml`:
    *  `/requirements.txt` : defines the dependencies of the project.
    *  `/package.json` : defines the packaging for YouWol ecosystem.
    *  `/$PROJECT_NAME/auto_generated.py`: defines runtime accessible information for the backend.
    (`$PROJECT_NAME` is the actual name of the project/package).
    """

    id = "setup"
    """
    ID of the step.

    Warning:
        The flows defined in this pipelines reference this ID, in common scenarios it should not be modified.
    """

    run: ExplicitNone = ExplicitNone()
    """
    The execution of the step is not provided as a shell command, rather it is implemented in the
    function `execute_run` of this class.
    """
    sources: FileListing = FileListing(include=[pyproject_file])
    """
    Source files of the step.

    Note:
        If those did not change since last execution, the step is considered in sync.
    """

    async def execute_run(self, project: Project, flow_id: str, context: Context):
        """
        Trigger step execution.

        Parameters:
            project: Project for which the step is executed.
            flow_id: ID of the flow associated.
            context: Current context.

        Return:
            Manifest of the execution
        """
        async with context.start(
            action="SetupStep",
        ):
            with open(project.path / pyproject_file, "rb") as f:
                pyproject = tomllib.load(f)

                package_json = SetupStep.__package_json(pyproject)
                write_json(package_json, project.path / "package.json")

                requirements = SetupStep.__requirements_txt(pyproject)
                (project.path / "requirements.txt").write_text(requirements)

                auto_generated = SetupStep.__auto_generated_py(pyproject)
                (project.path / project.name / "auto_generated.py").write_text(
                    auto_generated
                )

                return {
                    "package_json": package_json,
                    "requirements": requirements,
                    "auto_generated": auto_generated,
                }

    @staticmethod
    def __package_json(pyproject: AnyDict) -> AnyDict:
        project_name = pyproject["project"]["name"]
        return {
            "name": project_name,
            "version": pyproject["project"]["version"],
            "main": f"{project_name}/main.py",
        }

    @staticmethod
    def __requirements_txt(pyproject: AnyDict) -> str:
        return f"""# This file is autogenerated from {pyproject_file}
{'\n'.join(pyproject["project"]["dependencies"])}"""

    @staticmethod
    def __auto_generated_py(pyproject: AnyDict) -> str:
        return f"""default_port = {pyproject['youwol']['default-port']}\n"""


class DependenciesStep(PipelineStep):
    """
    Creates a virtual environment `venv` in the project's folder and
    installs the dependencies from the `requirements.txt` file in it.
    """

    id: str = "dependencies"
    """
    ID of the step.

    Warning:
        The flows defined in this pipelines reference this ID, in common scenarios it should not be modified.
    """

    run: ExplicitNone = ExplicitNone()

    sources: FileListing = FileListing(
        include=[
            "requirements.txt",
        ]
    )
    """
    Source files of the step.

    Note:
        If those did not change since last execution, the step is considered in sync.
    """

    async def execute_run(self, project: Project, flow_id: str, context: Context):
        """
        Trigger step execution.

        Parameters:
            project: Project for which the step is executed.
            flow_id: ID of the flow associated.
            context: Current context.

        Return:
            Manifest of the execution
        """

        # for now this is required: 'sudo apt install python3.12-venv'
        async with context.start(
            action="DependenciesStep",
        ):
            # this is all temporary until py-youwol#0.1.7 is released
            venv_path = project.path / "venv"
            if venv_path.exists():
                shutil.rmtree(venv_path)

            cmd = (
                f"( cd {str(project.path)} "
                f"&& python3 -m venv venv "
                f"&& . venv/bin/activate "
                f"&& pip install -r ./requirements.txt)"
            )

            await execute_shell_cmd(cmd=cmd, context=context)
            youwol_path = Path(youwol.__path__[0])
            print(youwol_path)

            site_packages = glob(f"{venv_path}/lib/*/site-packages")[0]
            (Path(site_packages) / "youwol.pth").write_text(str(youwol_path.parent))


async def run_command(project: Project, context: Context) -> str:
    """
    Construct the shell command to serve the backend on `localhost`:
    *  Activates the virtual environment from the `venv` folder created with the
    [DependenciesStep](@yw-nav-class:youwol.pipelines.pipeline_python_backend.pipeline.DependenciesStep).
    *  Runs `python` from the `main.py` file of the package, provides `yw_port` from the current
    [YouwolEnvironment](@yw-nav-class:youwol.app.environment.youwol_environment.YouwolEnvironment).

    Parameters:
        project: Current project.
        context: Current context

    Return:
        The shell command.
    """
    async with context.start("run_command") as ctx:
        env = await ctx.get("env", YouwolEnvironment)
        return f"(. venv/bin/activate && python {project.name}/main.py --yw_port={env.httpPort})"


class RunStep(PipelineStep):
    """
    Starts the service.
    """

    id: str = "run"
    """
    ID of the step.

    Warning:
        The flows defined in this pipelines reference this ID, in common scenarios it should not be modified.
    """

    run: RunImplicit = lambda _step, project, _flow, ctx: run_command(
        project=project, context=ctx
    )
    """
    Shell command that starts the service.

    See [run_command](@yw-nav-func:youwol.pipelines.pipeline_python_backend.pipeline.run_command).
    """


default_packaging_files = FileListing(
    include=[
        "*",
        "*/**",
    ],
    ignore=["cdn.zip", "./.*", ".*/*", "**/.*/*", "venv"],
)
"""
Definition of the project's files packaged for publication.
"""


class PackageStep(PipelineStep):
    """
    Creates the artifact `package`, the one that will be published in local or remote ecosystems,
    from a files listing.

    The files listing is provided by
    [default_packaging_files](@yw-nav-glob:youwol.pipelines.pipeline_python_backend.pipeline.default_packaging_files).
    """

    id: str = "package"
    """
    ID of the step.

    Warning:
        The flows defined in this pipelines reference this ID, in common scenarios it should not be modified.
    """

    run: str = "echo 'Nothing to execute for packaging'"

    sources: FileListing = default_packaging_files
    """
    Source files of the step.

    Note:
        If those did not change since last execution, the step is considered in sync.
    """

    artifacts: list[Artifact] = [Artifact(id="package", files=default_packaging_files)]
    """
    One artifact is defined, it is called 'package' and contains all the files of the project by default.
    """


class PipelineConfig(BaseModel):
    """
    Specifies the configuration of the pipeline
    """

    with_tags: list[str] = []
    """
    The list of tags.
    """


async def pipeline(config: PipelineConfig, context: Context):
    """
    Instantiates the pipeline.

    It includes the following steps:
    *  [SetupStep](@yw-nav-class:youwol.pipelines.pipeline_python_backend.pipeline.SetupStep):
    Initializes of auto-generated files.
    *  [DependenciesStep](@yw-nav-class:youwol.pipelines.pipeline_python_backend.pipeline.DependenciesStep):
    Creates & setups a virtual environment.
    *  [PackageStep](@yw-nav-class:youwol.pipelines.pipeline_python_backend.pipeline.PackageStep):
    Packages selected files into a single artifact.
    *  [PublishCdnLocalStep](@yw-nav-class:youwol.pipelines.publish_cdn.PublishCdnLocalStep):
    Publishes in the local ecosystem the artifact from the `PackageStep`.
    *  One or more [PublishCdnRemoteStep](@yw-nav-class:youwol.pipelines.publish_cdn.PublishCdnLocalStep):
    Publishes in remote ecosystem(s) the package from the local ecosystem.
    *  [RunStep](@yw-nav-class:youwol.pipelines.pipeline_python_backend.pipeline.RunStep):
    Runs the service.

    It includes the following flows:
    *  `prod`: `['setup > dependencies > package > cdn-local > *cdn-remote', 'dependencies > run']` where
    `*cdn-remote` are the steps publishing the package in remote ecosystem(s).

    Parameters:
        config: configuration of the pipeline
        context: current context

    Return:
        The pipeline instance.
    """
    async with context.start(action="pipeline") as ctx:
        await ctx.info(text="Instantiate pipeline", data=config)

        publish_remote_steps, dags = await create_sub_pipelines_publish_cdn(
            start_step="cdn-local", targets=get_environment().cdnTargets, context=ctx
        )
        dependencies_step = DependenciesStep()
        package_step = PackageStep()
        steps = [
            SetupStep(),
            dependencies_step,
            package_step,
            RunStep(),
            PublishCdnLocalStep(
                packagedArtifacts=["package"],
            ),
            *publish_remote_steps,
        ]

        def parse_toml(project_folder: Path):
            with open(project_folder / pyproject_file, "rb") as f:
                return tomllib.load(f)

        return Pipeline(
            target=Target(family=Family.service),
            tags=config.with_tags,
            projectName=lambda path: parse_toml(path)["project"]["name"],
            projectVersion=lambda path: parse_toml(path)["project"]["version"],
            steps=steps,
            flows=[
                Flow(
                    name="prod",
                    dag=[
                        "setup > dependencies > package > cdn-local",
                        *dags,
                        "dependencies > run",
                    ],
                )
            ],
        )
